import numpy as np
import glob
import os, sys
import collections
import argparse

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

from lightningcast import utils
from datetime import datetime, timedelta
import skimage.measure

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
# ------------------------------------------------------------------------------------------------------------
def aggregate(
    inpatt,
    start_time,
    end_time,
    accumulation_duration=60,
    every_X_min=5,
    outpattern=os.environ["PWD"] + "/OUTPUT/%Y%m%d/FED_accum_60min_2km/",
    reduce_res=False,
    abidir="/arcdata/goes/grb/goes16/",
    sector="RadC",
    shape=(1500, 2500),
):

    startdt = dt = datetime.strptime(start_time, "%Y%m%d-%H%M")
    enddt = datetime.strptime(end_time, "%Y%m%d-%H%M")

    while startdt <= dt <= enddt:
        try:
            globbed_file = glob.glob(
                dt.strftime(
                    f"{abidir}/%Y/%Y_%m_%d_%j/abi/L1b/{sector}/*C13*_s%Y%j%H%M*"
                )
            )[0]
        except IndexError:
            # increment dt. Checking every minute for ABI data.
            dt = dt + timedelta(minutes=1)
            continue

        accum_dt = dt + timedelta(
            minutes=5
        )  # add 5 min since the agg ltg files are 5-min aggregations, ending on the dt

        accum_end_dt = dt + timedelta(minutes=accumulation_duration)
        print("dt of first file:", accum_dt, "; dt of last file:", accum_end_dt)

        while accum_dt <= accum_end_dt:
            try:
                datafile = glob.glob(accum_dt.strftime(inpatt))[0]
            except IndexError:
                for mm in [1, 2, 3]:
                    tmpdt = accum_dt - timedelta(minutes=mm)
                    print(tmpdt.strftime(inpatt))
                    try:
                        datafile = glob.glob(tmpdt.strftime(inpatt))[0]
                    except IndexError:
                        if mm == 3:
                            print("Can't find times close to", accum_dt)
                            print(accum_dt.strftime(inpatt))
                            print("")
                            try:
                                del accum_grid  # delete so the try-except block after while loop throws NameError and no file is written.
                            except NameError:
                                pass
                            datafile = None
                    else:
                        break  # breaks for-loop
            if datafile is None:
                break  # breaks while-loop and moves on to next accum_dt

            if (
                shape[0] == 5424
            ):  # indicates full-disk file from UMD, with different name
                fedvar = "Flash_extent_density_window"
            else:
                fedvar = "flash_extent_density"
            data = {
                fedvar: {"atts": []},
                "x": {"atts": ["ALL"]},
                "y": {"atts": ["ALL"]},
                "goes_imager_projection": {"atts": ["ALL"]},
            }
            gatts = {
                "subsatellite_longitude": -1,
                "platform_ID": -1,
                "orbital_slot": -1,
                "spatial_resolution": -1,
            }
            s = utils.read_netcdf(datafile, datasets=data, global_atts=gatts)
            tmpgrid = np.expand_dims(data[fedvar]["data"], axis=0)
            if shape == (1500, 2500):  # if the default, use the native shape
                native_shape = data[fedvar]["data"].shape
                newshape = (1, native_shape[0], native_shape[1])
            else:  # if NOT default
                newshape = (1, shape[0], shape[1])
            grid = np.zeros(newshape)

            append = False
            try:  # this try-except-block is needed because some of our GLM data is off by one row and one column
                grid[0] = tmpgrid
            except ValueError:
                try:
                    grid[0, 0:1499, 0:2499] = tmpgrid
                except ValueError:
                    print("Error: incompatible shape,", tmpgrid.shape, datafile)
                    try:
                        del accum_grid  # delete so the try-except block after while loop throws NameError and no file is written.
                    except NameError:
                        pass
                    break  # break while-loop
                else:
                    # we also need to append a value to x and y
                    append = True

            try:
                accum_grid = np.concatenate((accum_grid, grid), axis=0)
            except NameError:
                accum_grid = grid + 0.0
            # increment accum_dt
            accum_dt = accum_dt + timedelta(minutes=every_X_min)
        # end inner while-loop

        # take maximum
        try:
            final_accum = np.max(accum_grid, axis=0).astype(np.float32)
        except NameError:
            pass
        else:
            if reduce_res:
                # this is max pooling...not sure why the strange name
                final_accum = skimage.measure.block_reduce(
                    final_accum, (4, 4), np.max
                )  # reduce from 2-km to 8-km resolution
                data["x"]["data"] = data["x"]["data"][
                    2::4
                ]  # use the center (x,y)-point of the each new max-pooled region
                data["y"]["data"] = data["y"]["data"][2::4]
                gatts["spatial_resolution"] = "reduced from 2-km to 8-km"

            # write out file
            outfile = accum_end_dt.strftime(
                os.path.join(outpattern, "%Y%m%d-%H%M00.netcdf")
            )
            y, x = final_accum.shape

            dims = collections.OrderedDict()
            dims["x"] = x
            dims["y"] = y
            dims["g"] = 1
            dataset = collections.OrderedDict()
            # scale the x and y properly and save as np.uint16
            try:
                xpacked = (
                    (data["x"]["data"] - data["x"]["atts"]["add_offset"])
                    / data["x"]["atts"]["scale_factor"]
                ).astype(np.uint16)
                ypacked = (
                    (data["y"]["data"] - data["y"]["atts"]["add_offset"])
                    / data["y"]["atts"]["scale_factor"]
                ).astype(np.uint16)
            except KeyError:  # this means that x and y in the agg files were saved as floats. Just hardcode the offsets/factors
                data["x"]["atts"]["add_offset"] = -0.151844
                data["x"]["atts"]["scale_factor"] = 5.6e-05
                data["y"]["atts"]["add_offset"] = 0.151844
                data["y"]["atts"]["scale_factor"] = -5.6e-05
                xpacked = (
                    (data["x"]["data"] - data["x"]["atts"]["add_offset"])
                    / data["x"]["atts"]["scale_factor"]
                ).astype(np.uint16)
                ypacked = (
                    (data["y"]["data"] - data["y"]["atts"]["add_offset"])
                    / data["y"]["atts"]["scale_factor"]
                ).astype(np.uint16)
            if append:
                # make x and y and fed data have compatiable dims.
                # Hardcoding the values we need. Only works for GOES-East FGF CONUS sector
                xpacked = np.concatenate(([902], xpacked)).astype(np.uint16)
                ypacked = np.concatenate((ypacked, [1921])).astype(np.uint16)

            dataset["x"] = {"data": xpacked, "dims": ("x"), "atts": data["x"]["atts"]}
            dataset["y"] = {"data": ypacked, "dims": ("y"), "atts": data["y"]["atts"]}
            dataset["goes_imager_projection"] = {
                "data": data["goes_imager_projection"]["data"],
                "dims": ("g"),
                "atts": data["goes_imager_projection"]["atts"],
            }
            varname = (
                os.path.basename(outpattern[:-1])
                if (outpattern[-1] == "/")
                else os.path.basename(outpattern)
            )
            dataset[varname] = {
                "data": final_accum,
                "dims": ("y", "x"),
                "atts": {
                    "units": "flashes",
                    "long_name": "Max flash-extent density in previous "
                    + str(accumulation_duration)
                    + " min",
                },
            }
            gatts["sector"] = sector
            gatts["time_coverage_start"] = dt.strftime("%Y-%m-%dT%H:%M:%SZ")
            gatts["time_coverage_end"] = accum_end_dt.strftime("%Y-%m-%dT%H:%M:%SZ")
            gatts["aggregation"] = str(accumulation_duration) + " min"

            status = utils.write_netcdf(
                outfile, dataset, dims, gatts, wait=False, gzip=False, nc4=True
            )
            print("wrote out " + outfile)
            print("")
            del accum_grid, final_accum

        # increment dt. Checking every minute for ABI data.
        dt = dt + timedelta(minutes=1)


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "inpatt",
        help="Pattern for the data. E.g., /apollo/grain/saved_probsevere_data/lightning/%%Y/%%Y%%m%%d/GLM/goes_east/agg/%%Y%%m%%d-%%H%%M*.netcdf",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "start_time",
        help="datetime to begin accumulation. E.g., 20201001-2021",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "end_time",
        help="datetime to end accumulation. E.g., 20201011-0342",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-i",
        "--inabidir",
        help="ABI dir to search for patterns to match. Default='/arcdata/goes/grb/goes16/'",
        default=["/arcdata/goes/grb/goes16/"],
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-a",
        "--accumulation_duration",
        help="How long to accumulate in minutes. Default=60.",
        nargs=1,
        default=[60],
        type=int,
    )
    # parser.add_argument('-e','--every_X_min',help="Every X minutes to do the accumulation. Default=5.",nargs=1,default=[5],type=int)
    parser.add_argument(
        "-o",
        "--outpattern",
        help="Output directory pattern. Default=${PWD}/OUTPUT/%%Y%%m%%d/FED_accum_60min_2km/",
        nargs=1,
        default=[os.environ["PWD"] + "/OUTPUT/%Y%m%d/FED_accum_60min_2km/"],
        type=str,
    )
    parser.add_argument(
        "-s",
        "--sector",
        help="Sector to use. Default = RadC",
        default=["RadC"],
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-sh",
        "--shape",
        help="Shape of the output data (using 2km). Default = 1500 2500 if RadC. \
                      args.shape if otherwise.",
        nargs=2,
        type=int,
    )
    parser.add_argument(
        "-r",
        "--reduce_res",
        help="Reduce resolution from 2-km (default) to 8-km.",
        action="store_true",
    )

    args = parser.parse_args()

    if args.shape:
        shape = (args.shape[0], args.shape[1])
    elif args.sector[0] == "RadC":
        shape = (1500, 2500)
    # else:
    #  shape = args.shape #(2500,3200)
    # shape = args.shape

    aggregate(
        args.inpatt[0],
        args.start_time[0],
        args.end_time[0],
        abidir=args.inabidir[0],
        accumulation_duration=args.accumulation_duration[0],
        every_X_min=5,  # currently static
        outpattern=args.outpattern[0],
        reduce_res=args.reduce_res,
        sector=args.sector[0],
        shape=shape,
    )
    # example run for FD data from UMD
    # python accumulation.py /ships19/grain/jcintineo/GLM/goes_east/FD/%Y/%Y%m%d/5min/*e%Y%m%d%H%M*.nc 20200505-0001 20200505-0100 -s RadF -sh 5424 5424 -o /ships19/grain/jcintineo/GLM/goes_east/FD/%Y/%Y%m%d/FED_accum_60min/
