import os
import glob
import sys

pltg = os.environ["PLTG"]
import argparse
import numpy as np
import time
from datetime import datetime, timedelta
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
import shutil
from lightningcast import utils
import pygrib as pg
import pyresample as pr
from subprocess import Popen, call, PIPE

# GLOBALS:
new_grid, new_ny, new_nx = utils.get_area_definition(pltg + "/src/static/GOES_East.nc")


class rap_cycle:
    def __init__(self, date, cycle):
        self.date = date
        self.cycle = cycle
        self.name = date + "-" + cycle
        self.age = 0.0


class hrrr_cycle:
    def __init__(self, date, cycle):
        self.date = date
        self.cycle = cycle
        self.name = date + "-" + cycle
        self.age = 0.0


################################################################################################################
def remap_hrrr(grib2, outroot):
    logging.info("Process started")

    gr = pg.open(grib2)
    lats = gr[1]["latitudes"]
    lons = gr[1]["longitudes"] - 360
    nx = gr[1]["Nx"]
    ny = gr[1]["Ny"]
    lats = np.reshape(lats, (ny, nx))
    lons = np.reshape(lons, (ny, nx))
    old_grid = pr.geometry.GridDefinition(lons=lons, lats=lats)

    analDate = gr[1].analDate
    validDate = gr[1].validDate
    strdate = str(analDate).replace("-", "")[0:8]  # yyyymmdd

    # get grids we want
    u925 = (
        gr.select(name="U component of wind", level=925)[0].values * 1.94384
    )  # ==convert to kts
    v925 = (
        gr.select(name="V component of wind", level=925)[0].values * 1.94384
    )  # ==convert to kts
    u850 = (
        gr.select(name="U component of wind", level=850)[0].values * 1.94384
    )  # ==convert to kts
    v850 = (
        gr.select(name="V component of wind", level=850)[0].values * 1.94384
    )  # ==convert to kts
    u700 = (
        gr.select(name="U component of wind", level=700)[0].values * 1.94384
    )  # ==convert to kts
    v700 = (
        gr.select(name="V component of wind", level=700)[0].values * 1.94384
    )  # ==convert to kts
    u500 = (
        gr.select(name="U component of wind", level=500)[0].values * 1.94384
    )  # ==convert to kts
    v500 = (
        gr.select(name="V component of wind", level=500)[0].values * 1.94384
    )  # ==convert to kts
    mlcape = gr.select(name="Convective available potential energy", level=9000)[
        0
    ].values
    # land_sea_mask = gr.select(name='Land-sea mask',level=0)[0].values
    gr.close()

    # remap the grids
    roi = 10000
    r_u925 = pr.kd_tree.resample_nearest(
        old_grid, u925, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_v925 = pr.kd_tree.resample_nearest(
        old_grid, v925, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_u850 = pr.kd_tree.resample_nearest(
        old_grid, u850, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_v850 = pr.kd_tree.resample_nearest(
        old_grid, v850, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_u700 = pr.kd_tree.resample_nearest(
        old_grid, u700, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_v700 = pr.kd_tree.resample_nearest(
        old_grid, v700, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_u500 = pr.kd_tree.resample_nearest(
        old_grid, u500, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_v500 = pr.kd_tree.resample_nearest(
        old_grid, v500, new_grid, radius_of_influence=roi, fill_value=-999
    )
    r_mlcape = pr.kd_tree.resample_nearest(
        old_grid, mlcape, new_grid, radius_of_influence=roi, fill_value=-999
    )

    # write data
    dims = {}
    dims["x"] = new_nx
    dims["y"] = new_ny

    gatts = {}
    gatts["cycle"] = str(analDate)
    gatts["forecast_hour"] = str(validDate)

    datasets = {}
    datasets["u925"] = {"data": r_u925, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["v925"] = {"data": r_v925, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["u850"] = {"data": r_u850, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["v850"] = {"data": r_v850, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["u700"] = {"data": r_u700, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["v700"] = {"data": r_v700, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["u500"] = {"data": r_u500, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["v500"] = {"data": r_v500, "atts": {"units": "kts"}, "dims": ("y", "x")}
    datasets["mlcape"] = {
        "data": r_mlcape,
        "atts": {"units": "J/km", "description": "0-90mb AGL avg."},
        "dims": ("y", "x"),
    }

    filename = (
        "HRRR_remapped_"
        + (str(analDate)).replace("-", "").replace(":", "").replace(" ", "_")
        + ".nc"
    )
    outfile = os.path.join(outroot, strdate, filename)
    status = utils.write_netcdf(outfile, datasets, dims, atts=gatts)

    logging.info("Process ended")


################################################################################################################
def main():
    pltg_data = os.environ["PLTG_DATA"] + "/"
    nwp_data = pltg_data + "NWP_RAW"
    nwp_grib2_dir = nwp_data + "/GRIB2/"

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-burl",
        "--baseurl",
        help="[optional] e.g., 'http://nomads.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/'",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-bdir",
        "--basedir",
        help="[optional] e.g.,'/common/data/model/com/rap/prod/'",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-s",
        "--sleepInterval",
        help="[optional] Default = 300 sec [5 min]",
        default=[300],
        type=int,
        nargs=1,
    )
    parser.add_argument(
        "-ph",
        "--prevHours",
        help="[optional] Default = 6 hours. Number of previous hours to process RAP data",
        default=[6],
        type=int,
        nargs=1,
    )
    parser.add_argument(
        "-l",
        "--logfile",
        help="[optional] Default: pltg_data + '/logs/'+model+'.log'",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-gt",
        "--geotiff",
        help="[optional] Set this flag to make geotiffs of certain NWP products and send to RE VM.",
        action="store_true",
    )

    args = parser.parse_args()

    if not (args.baseurl) and not (args.basedir):
        logging.error("You must provide either a baseurl or a basedir")
        sys.exit(1)
    else:
        if args.baseurl:
            baseurl = args.baseurl[0]
            if baseurl[-1] != "/":
                baseurl += "/"
            basedir = None
            basePath = baseurl
        else:
            basedir = args.basedir[0]
            if basedir[-1] != "/":
                basedir += "/"
            baseurl = None
            basePath = basedir

    if "hrrr" in basePath:
        model = "hrrr"
    if "rap" in basePath:
        model = "rap"
    if "gfs" in basePath:
        model = "gfs"

    if not (args.logfile):
        logfile = pltg_data + "/logs/" + model + ".log"
    else:
        logfile = args.logfile[0]

    # define the logger
    logDict = logging_def(logfile=logfile)
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    logging.info("Process started")

    sleepInterval = args.sleepInterval[0]
    if sleepInterval < 30:
        logging.warning("You are probably polling too frequently.")
    if sleepInterval > 60 * 60:
        logging.warning("You are probably polling too infrequently.")

    prevHours = args.prevHours[0]
    if prevHours < 3:
        logging.warning(
            "You may want to process more than " + str(prevHours) + " previous hours."
        )
    if prevHours > 24:
        logging.warning(
            "You probably do not want to process more than "
            + str(prevHours)
            + " previous hours."
        )

    logging.info("Listening to " + basePath + " for " + model.upper() + " data.")

    while True:
        now = datetime.now()
        nowsecs = int(time.time())
        wait_min = 10
        wait_time_thresh = nowsecs - 60 * wait_min
        year = now.strftime("%Y")
        month = now.strftime("%m")
        day = now.strftime("%d")
        full_date = year + month + day

        if not (os.path.isdir(nwp_grib2_dir)):
            os.makedirs(nwp_grib2_dir)

        if model == "gfs":
            gfs_hdf_dir = nwp_data + "/GFS_HDF/"
            if not (os.path.isdir(gfs_hdf_dir)):
                os.makedirs(gfs_hdf_dir)
            convert_dir = probsev + "/src/convert_gfs/grib2hdf/"
            wgrib2_dir = os.environ["WGRIB2"]
            gfs_cycles = ["00", "06", "12", "18"]

            for cycle in gfs_cycles:
                next_grib2_name = "gfs.t" + cycle + "z.pgrb2.0p50.f012"
                better_grib2_name = (
                    "gfs." + full_date + ".t" + cycle + "zpgrb2.0p50.f012"
                )
                expected_hdf = (
                    gfs_hdf_dir + "gfs." + full_date[2:] + cycle + "_F012.hdf"
                )
                # check to see if grib2 file exists locally
                if not (os.path.isfile(nwp_grib2_dir + better_grib2_name)):
                    # doesn't exist, so fetch and try to process
                    if basedir:
                        # check to see if grib2 exists on mounted disk
                        mounted_file = (
                            basedir + "gfs." + full_date + cycle + "/" + next_grib2_name
                        )
                        if os.path.isfile(mounted_file):
                            try:
                                shutil.copy(
                                    mounted_file, nwp_grib2_dir + better_grib2_name
                                )
                            except (FileNotFoundError, PermissionError) as err:
                                logging.error(str(err))
                                break
                            else:
                                logging.info(
                                    "grib2hdf on " + nwp_grib2_dir + better_grib2_name
                                )
                                proc = Popen(
                                    [
                                        convert_dir + "convert_grib_hdf",
                                        "2",
                                        "ZXY",
                                        better_grib2_name,
                                        wgrib2_dir,
                                        nwp_grib2_dir,
                                        gfs_hdf_dir,
                                    ],
                                    stdout=PIPE,
                                    stderr=PIPE,
                                )
                                stdout, stderr = proc.communicate()
                                logging.info("grib2hdf done")
                                if not (os.path.isfile(expected_hdf)):
                                    logging.warning(
                                        expected_hdf
                                        + " does not exist. Probable grib2hdf failure."
                                    )

                    else:  # baseurl
                        # check to see if grib2 exists on http
                        http_file = (
                            baseurl
                            + "gfs."
                            + year
                            + month
                            + day
                            + "/"
                            + cycle
                            + "/"
                            + next_grib2_name
                        )
                        proc = Popen(
                            [
                                "wget",
                                http_file,
                                "-O",
                                nwp_grib2_dir + better_grib2_name,
                            ],
                            stdout=PIPE,
                            stderr=PIPE,
                        )
                        stdout, stderr = proc.communicate()
                        if not (isinstance(stdout, str)):
                            stdout = stdout.decode("utf-8")
                        if not (isinstance(stderr, str)):
                            stderr = stderr.decode("utf-8")
                        if "failed" in stderr.lower() or "error" in stderr.lower():
                            logging.error(stderr)
                            if os.path.isfile(nwp_grib2_dir + better_grib2_name):
                                os.remove(nwp_grib2_dir + better_grib2_name)
                        elif "not found" in stderr.lower():
                            logging.info(http_file + " not available yet.")
                            if os.path.isfile(nwp_grib2_dir + better_grib2_name):
                                os.remove(nwp_grib2_dir + better_grib2_name)
                        else:
                            logging.info(
                                "grib2hdf on " + nwp_grib2_dir + better_grib2_name
                            )
                            proc = Popen(
                                [
                                    convert_dir + "convert_grib_hdf",
                                    "2",
                                    "ZXY",
                                    better_grib2_name,
                                    wgrib2_dir,
                                    nwp_grib2_dir,
                                    gfs_hdf_dir,
                                ],
                                stdout=PIPE,
                                stderr=PIPE,
                            )
                            stdout, stderr = proc.communicate()
                            logging.info("grib2hdf done")
                            if not (os.path.isfile(expected_hdf)):
                                logging.warning(
                                    expected_hdf
                                    + " does not exist. Probable grib2hdf failure."
                                )

        if model == "rap":
            rap_forecast_hours = ["00", "01", "02", "03"]
            num_good_fhs = len(rap_forecast_hours)
            rap_cycles = []
            for cc in range(prevHours + 1):
                prev_dt = now - timedelta(hours=cc)
                rco = rap_cycle(prev_dt.strftime("%Y%m%d"), prev_dt.strftime("%H"))
                rap_cycles.append(rco)

            for ii in range(len(rap_cycles)):
                # finished file is the final file form we expect (e.g., for EBS)
                finished_file = glob.glob(
                    ("/").join(
                        [
                            probsev_data,
                            "NWP_data",
                            rap_cycles[ii].date,
                            "EBS_merged_smoothed",
                            "00.00",
                            rap_cycles[ii].date
                            + "-"
                            + rap_cycles[ii].cycle
                            + "0000.netcdf*",
                        ]
                    )
                )
                if len(finished_file) == 0:  # doesn't exist yet
                    for fh in rap_forecast_hours:
                        # check to see if grib2 exists locally
                        full_file = (
                            nwp_grib2_dir
                            + rap_cycles[ii].date
                            + "/rap.t"
                            + rap_cycles[ii].cycle
                            + "z.awp130pgrbf"
                            + fh
                            + ".grib2"
                        )

                        # mkdir first, if it doesn't exist
                        dirname = os.path.dirname(full_file)
                        if not (os.path.isdir(dirname)):
                            os.makedirs(dirname)

                        if not (os.path.isfile(full_file)):
                            # doesn't exist so fetch
                            if basedir:
                                # check to see if grib2 exists on mounted disk
                                mounted_file = (
                                    basedir
                                    + "rap."
                                    + rap_cycles[ii].date
                                    + "/rap.t"
                                    + rap_cycles[ii].cycle
                                    + "z.awp130pgrbf"
                                    + fh
                                    + ".grib2"
                                )
                                if os.path.isfile(mounted_file):
                                    try:
                                        shutil.copy(mounted_file, full_file)
                                    except (FileNotFoundError, PermissionError) as err:
                                        logging.error(str(err))
                                        break
                                    else:
                                        logging.info("Got " + mounted_file)
                            else:  # baseurl
                                # check to see if grib2 exists on http
                                http_file = (
                                    baseurl
                                    + "rap."
                                    + rap_cycles[ii].date
                                    + "/rap.t"
                                    + rap_cycles[ii].cycle
                                    + "z.awp130pgrbf"
                                    + fh
                                    + ".grib2"
                                )
                                proc = Popen(
                                    ["wget", http_file, "-O", full_file, "-t", "1"],
                                    stdout=PIPE,
                                    stderr=PIPE,
                                )
                                stdout, stderr = proc.communicate()
                                if not (isinstance(stdout, str)):
                                    stdout = stdout.decode("utf-8")
                                if not (isinstance(stderr, str)):
                                    stderr = stderr.decode("utf-8")
                                if (
                                    "failed" in stderr.lower()
                                    or "error" in stderr.lower()
                                ):
                                    logging.error(stderr)
                                    if os.path.isfile(full_file):
                                        os.remove(full_file)
                                elif "not found" in stderr.lower():
                                    logging.info(http_file + " not available yet.")
                                    if os.path.isfile(full_file):
                                        os.remove(full_file)
                                else:
                                    logging.info("Got " + http_file)

                        # Does the expected output of createFinalNseFields_RT with calculate_fields=True exist?
                        expected_file = ("/").join(
                            [
                                probsev_data,
                                "NWP_data",
                                rap_cycles[ii].date,
                                "RAP",
                                fh,
                                "EBS",
                                rap_cycles[ii].date
                                + "-"
                                + rap_cycles[ii].cycle
                                + "0000.netcdf",
                            ]
                        )
                        if not (os.path.isfile(expected_file)) and os.path.isfile(
                            full_file
                        ):  # check grib2 again, just in case wget/cp failed
                            createFinalNseFields_RT(
                                [int(rap_cycles[ii].date)],
                                [rap_cycles[ii].cycle],
                                [fh],
                                calculate_fields=True,
                            )

                    # we have not yet created final grids for this cycle. Complete processing
                    processed_forecast_grids = glob.glob(
                        ("/").join(
                            [
                                probsev_data,
                                "NWP_data",
                                rap_cycles[ii].date,
                                "RAP",
                                "*",
                                "EBS",
                                rap_cycles[ii].date
                                + "-"
                                + rap_cycles[ii].cycle
                                + "0000.netcdf*",
                            ]
                        )
                    )
                    num_fhs = len(processed_forecast_grids)
                    if num_fhs == num_good_fhs:
                        logging.info(
                            "We have all "
                            + str(num_fhs)
                            + " processed grids for "
                            + rap_cycles[ii].date
                            + "-"
                            + rap_cycles[ii].cycle
                            + ". Finish processing."
                        )
                        # if we have all the fhs pre-processed (i.e., grids made), go ahead and complete processing
                        createFinalNseFields_RT(
                            [int(rap_cycles[ii].date)],
                            [rap_cycles[ii].cycle],
                            ["00"],
                            merge_fields=True,
                            smooth=True,
                            realtime=True,
                            geotiff=args.geotiff,
                        )  # FCST_HOURS does not matter
                    if num_fhs == (num_good_fhs - 1):
                        # check age of first processed_forecast_grid, if older than 10 min, then go ahead and process with 1 missing grid.
                        mtime = os.path.getmtime(processed_forecast_grids[0])
                        if mtime < wait_time_thresh:
                            logging.info(
                                "We have "
                                + str(num_fhs)
                                + "/"
                                + str(num_good_fhs)
                                + " processed grids for "
                                + rap_cycles[ii].date
                                + "-"
                                + rap_cycles[ii].cycle
                                + ", but have waited "
                                + str(wait_min)
                                + " min. Finish processing."
                            )
                            createFinalNseFields_RT(
                                [int(rap_cycles[ii].date)],
                                [rap_cycles[ii].cycle],
                                ["00"],
                                merge_fields=True,
                                smooth=True,
                                realtime=True,
                                geotiff=args.geotiff,
                            )  # FCST_HOURS does not matter

                else:
                    logging.info("We already have " + finished_file[0])

        if model == "hrrr":
            hrrr_forecast_hours = ["01"]
            num_good_fhs = len(hrrr_forecast_hours)
            hrrr_cycles = []
            for cc in range(prevHours + 1):
                prev_dt = now - timedelta(hours=cc)
                rco = hrrr_cycle(prev_dt.strftime("%Y%m%d"), prev_dt.strftime("%H"))
                hrrr_cycles.append(rco)

            for ii in range(len(hrrr_cycles)):
                # finished file is the final file form we expect (e.g., for EBS)
                finished_file = glob.glob(
                    os.path.join(
                        pltg_data,
                        "NWP",
                        hrrr_cycles[ii].date,
                        "HRRR_remapped_"
                        + hrrr_cycles[ii].date
                        + "+"
                        + hrrr_cycles[ii].cycle
                        + "0000.nc*",
                    )
                )
                if len(finished_file) == 0:  # doesn't exist yet
                    for fh in hrrr_forecast_hours:
                        # check to see if grib2 exists locally
                        full_file = (
                            nwp_grib2_dir
                            + hrrr_cycles[ii].date
                            + "/hrrr.t"
                            + hrrr_cycles[ii].cycle
                            + "z.wrfsfcf"
                            + fh
                            + ".grib2"
                        )

                        # mkdir first, if it doesn't exist
                        dirname = os.path.dirname(full_file)
                        if not (os.path.isdir(dirname)):
                            os.makedirs(dirname)

                        if not (os.path.isfile(full_file)):
                            # doesn't exist so fetch
                            if basedir:
                                # check to see if grib2 exists on mounted disk
                                mounted_file = (
                                    basedir
                                    + "hrrr."
                                    + hrrr_cycles[ii].date
                                    + "/conus/hrrr.t"
                                    + hrrr_cycles[ii].cycle
                                    + "z.wrfsfcf"
                                    + fh
                                    + ".grib2"
                                )
                                if os.path.isfile(mounted_file):
                                    try:
                                        shutil.copy(mounted_file, full_file)
                                    except (FileNotFoundError, PermissionError) as err:
                                        logging.error(str(err))
                                        break
                                    else:
                                        logging.info("Got " + mounted_file)
                            else:  # baseurl
                                # check to see if grib2 exists on http
                                http_file = (
                                    baseurl
                                    + "hrrr."
                                    + hrrr_cycles[ii].date
                                    + "/conus/hrrr.t"
                                    + hrrr_cycles[ii].cycle
                                    + "z.wrfsfcf"
                                    + fh
                                    + ".grib2"
                                )
                                proc = Popen(
                                    ["wget", http_file, "-O", full_file, "-t", "1"],
                                    stdout=PIPE,
                                    stderr=PIPE,
                                )
                                stdout, stderr = proc.communicate()
                                if not (isinstance(stdout, str)):
                                    stdout = stdout.decode("utf-8")
                                if not (isinstance(stderr, str)):
                                    stderr = stderr.decode("utf-8")
                                if (
                                    "failed" in stderr.lower()
                                    or "error" in stderr.lower()
                                ):
                                    logging.error(stderr)
                                    if os.path.isfile(full_file):
                                        os.remove(full_file)
                                elif "not found" in stderr.lower():
                                    logging.info(http_file + " not available yet.")
                                    if os.path.isfile(full_file):
                                        os.remove(full_file)
                                else:
                                    logging.info("Got " + http_file)

                        # Does the expected output of createFinalNseFields_RT with calculate_fields=True exist?
                        expected_file = os.path.join(
                            pltg_data,
                            "NWP",
                            hrrr_cycles[ii].date,
                            "HRRR_remapped_"
                            + hrrr_cycles[ii].date
                            + "_"
                            + hrrr_cycles[ii].cycle
                            + "0000.nc",
                        )
                        ef2 = expected_file + ".gz"
                        print(expected_file)
                        if (
                            not (os.path.isfile(expected_file))
                            and not (os.path.isfile(ef2))
                            and os.path.isfile(full_file)
                        ):  # check grib2 again, just in case wget/cp failed
                            remap_hrrr(full_file, os.path.join(pltg_data, "NWP"))
                        else:
                            logging.info(
                                expected_file
                                + " already exists OR "
                                + full_file
                                + " does not."
                            )

        logging.info("sleeping " + str(sleepInterval) + " seconds.")
        time.sleep(sleepInterval)


if __name__ == "__main__":
    main()
