from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import numpy as np
from lightningcast import ams_utils
import sklearn
import pickle
import argparse
import shutil

import tensorflow as tf

tf.keras.backend.clear_session()
from tensorflow import keras
from tensorflow.keras.models import Model, save_model, load_model
from lightningcast import keras_metrics

from lightningcast import performance_diagrams
from lightningcast.tf_eval import get_area_under_perf_diagram
from lightningcast import tf_generator

PERMUTED_PREDICTORS_KEY = "permuted_predictor_name_by_step"
HIGHEST_COSTS_KEY = "highest_cost_by_step"
ORIGINAL_COST_KEY = "original_cost"
STEP1_PREDICTORS_KEY = "predictor_names_step1"
STEP1_COSTS_KEY = "costs_step1"


def negative_aupd(preds, labels):

    scores_dict = performance_diagrams._get_points_in_perf_diagram(
        observed_labels=labels, forecast_probabilities=preds
    )
    pod_by_threshold = scores_dict["pod"]
    success_ratio_by_threshold = scores_dict["sr"]

    aupd = get_area_under_perf_diagram(success_ratio_by_threshold, pod_by_threshold)

    return aupd * -1


def make_preds(conv_model, config, DG, val_list, permute_input=[]):
    stride = 4

    config["shuffle"] = False
    config["batch_size"] = 1
    validation_generator = DG(val_list, config, permute_input=permute_input)
    n_vsamples = len(val_list)  # from lustre/samplelists only
    ny = val_list[0]["endY"] - val_list[0]["startY"]
    nx = val_list[0]["endX"] - val_list[0]["startX"]

    # preallocate
    all_preds = np.empty(
        (n_vsamples, ny // stride, nx // stride), dtype=np.float32
    )  # Striding to save space and time.
    all_labels = np.empty(
        (n_vsamples, ny // stride, nx // stride), dtype=np.float32
    )  # However, we don't stride for spatial_verification.
    # float32 precision needed for validation routines (e.g., MSE)

    cter = 0
    for inputs, labels in validation_generator:
        preds = conv_model.predict(inputs, workers=10, verbose=1)
        all_preds[cter] = preds[0, ::stride, ::stride]
        all_labels[cter] = labels[0, ::stride, ::stride]

        cter += 1

    all_preds = all_preds.flatten()
    all_labels = all_labels.flatten()

    return all_preds, all_labels


def permutation_test_for_cnn(
    cnn_model_object, config, samplelist_object, outdir, cost_function=negative_aupd
):
    """Runs permutation test on CNN (convolutional neural net).
    E = number of examples (storm objects)
    C = number of channels (predictor variables)
    :param cnn_model_object: Trained instance of `keras.models.Model`.
    :param config: dictionary for the trained model
    :param samplelist_object: a list of samples (dicts), containing lustre filesnames, for example
    :param outdir: Path to output file.  `result_dict` (the
        output variable) will be saved here.
    :param cost_function: Cost function (used to evaluate model predictions).
        Must be negatively oriented (lower values are better).  Must have the
        following inputs and outputs.

    :return: result_dict: Dictionary with the following keys.
    result_dict['permuted_predictor_name_by_step']: length-C list with name of
        predictor permuted at each step.
    result_dict['highest_cost_by_step']: length-C numpy array with corresponding
        cost at each step.  highest_cost_by_step[m] = cost after permuting
        permuted_predictor_name_by_step[m].
    result_dict['original_cost']: Original cost (before any permutation).
    result_dict['predictor_names_step1']: length-C list of predictor names.
    result_dict['costs_step1']: length-C numpy array of corresponding costs.
        costs_step1[m] = cost after permuting only predictor_names_step1[m].
        This key and "predictor_names_step1" correspond to the Breiman version
        of the permutation test, while "permuted_predictor_name_by_step" and
        "highest_cost_by_step" correspond to the Lakshmanan version.
    """

    scalar_vars = config["scalar_vars"]
    inputs = config["inputs"]  # non-scalar inputs

    if "LUSTRE" in config["datagen"].upper():
        DG = getattr(tf_generator, "DataGeneratorLUSTREPermutation")
    else:
        print("Permutation test not supported for " + config["datagen"])
        sys.exit(1)
    val_list = samplelist_object + []

    # This block generates a list of variable names
    # as well as a list of 2-elem tuples. The first element of each tuple is the input index (0,1,2, etc.)
    # and the second element of each tuples is the channel index for that input.
    permute_inputs = []
    predictor_names = []
    for ii, inp in enumerate(inputs):
        for jj, chan in enumerate(inp):
            predictor_names.append(chan)
            permute_inputs.append((ii, jj))
    for sv in range(len(scalar_vars)):
        predictor_names.append(scalar_vars[sv])
        permute_inputs.append((input_cter, sv))

    permanent_permute_inputs = []

    print("--------------------------------------------")

    remaining_predictor_names = predictor_names + []
    current_step_num = 0

    permuted_predictor_name_by_step = []
    highest_cost_by_step = []
    predictor_names_step1 = []
    costs_step1 = []

    # Get original cost (before permutation).
    these_probabilities, target_values = make_preds(
        cnn_model_object, config, DG, val_list
    )

    original_cost = cost_function(these_probabilities, target_values)
    print("Original cost (no permutation): {0:.4e}\n".format(original_cost))

    while len(remaining_predictor_names) > 0:
        current_step_num += 1

        highest_cost = -np.inf
        best_predictor_name = None
        best_predictor_permuted_values = None

        for this_predictor_name in remaining_predictor_names:
            message_string = (
                'Trying predictor "{0:s}" at step {1:d} of permutation test...' "\n"
            ).format(this_predictor_name, current_step_num)
            print(message_string)

            # new val generator
            curr_pis = permanent_permute_inputs.copy()
            curr_pis.append(this_predictor_name)

            these_probabilities, target_values = make_preds(
                cnn_model_object, config, DG, val_list, permute_input=curr_pis
            )

            this_cost = cost_function(these_probabilities, target_values)
            print("\nResulting cost = {0:.4e}".format(this_cost))
            print("--------------------------------------------")

            if current_step_num == 1:
                predictor_names_step1.append(this_predictor_name)
                costs_step1.append(this_cost)

            if this_cost < highest_cost:
                continue

            highest_cost = this_cost + 0.0
            best_predictor_name = this_predictor_name + ""
            # best_predictor_permuted_values = this_predictor_matrix[
            #    ..., this_predictor_index]

        permuted_predictor_name_by_step.append(best_predictor_name)
        highest_cost_by_step.append(highest_cost)

        # Remove best predictor from list.
        remaining_predictor_names.remove(best_predictor_name)

        # Leave values of best predictor permuted.
        # This means
        permanent_permute_inputs.append(best_predictor_name)

        print(
            '\nBest predictor = "{0:s}" ... new cost = {1:.4e}\n'.format(
                best_predictor_name, highest_cost
            )
        )

    result_dict = {
        PERMUTED_PREDICTORS_KEY: permuted_predictor_name_by_step,
        HIGHEST_COSTS_KEY: np.array(highest_cost_by_step),
        ORIGINAL_COST_KEY: original_cost,
        STEP1_PREDICTORS_KEY: predictor_names_step1,
        STEP1_COSTS_KEY: np.array(costs_step1),
    }

    output_pickle_file_name = f"{outdir}/permutation_results.pkl"

    ams_utils._create_directory(file_name=output_pickle_file_name)

    print('Writing results to: "{0:s}"...'.format(output_pickle_file_name))
    file_handle = open(output_pickle_file_name, "wb")
    pickle.dump(result_dict, file_handle)
    file_handle.close()

    return result_dict


#####################################################################################################################

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("model", help="Model to permute.", type=str, nargs=1)
    parser.add_argument("model_config", help="Associated config pkl", nargs=1, type=str)
    parser.add_argument(
        "samplelist",
        help="A samplelist (see create_lustre_samplelist.py). Usually a subset of validation data.",
        nargs=1,
        type=str,
    )
    parser.add_argument("outdir", help="Output directory.", nargs=1, type=str)

    args = parser.parse_args()

    outdir = args.outdir[0]

    model_object = load_model(args.model[0])
    config = pickle.load(open(args.model_config[0], "rb"))
    samplelist_object = pickle.load(open(args.samplelist[0], "rb"))[
        0::2
    ]  # doing 1/2 of the samplelist!

    permutation_dict = permutation_test_for_cnn(
        model_object, config, samplelist_object, outdir
    )
    # permutation_dict = pickle.load(open(os.path.join(indir,'permutation_test.pkl'),'rb'))

    breiman_file_name = "{0:s}/breiman_results.jpg".format(outdir)
    ams_utils.plot_breiman_results(
        result_dict=permutation_dict,
        output_file_name=breiman_file_name,
        plot_percent_increase=False,
    )

    lakshmanan_file_name = "{0:s}/lakshmanan_results.jpg".format(outdir)
    ams_utils.plot_lakshmanan_results(
        result_dict=permutation_dict,
        output_file_name=lakshmanan_file_name,
        plot_percent_increase=False,
    )

    shutil.copy(args.samplelist[0], outdir + "/")
