import os, sys
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import time

# -------------------------------------------------------------------------------------------------#
class DataGenerator(torch.utils.data.Dataset):
    "Characterizes a dataset for PyTorch"

    def __init__(self, nsamples):
        "Initialization"
        self.nsamples = nsamples

    def __len__(self):
        "Denotes the total number of samples"
        return self.nsamples

    def __getitem__(self, index):
        "Generate one sample of data"

        inputs = np.random.random((3, 32, 32))  # 3 channels; nx = ny = 32

        # targets
        y = np.random.random(10)  # 10 outputs

        return inputs, y


# -------------------------------------------------------------------------------------------------#
class Print(nn.Module):
    def forward(self, x):
        print(x.size())
        return x


# -------------------------------------------------------------------------------------------------#
def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group["lr"]


# -------------------------------------------------------------------------------------------------#
def count_parameters(model):
    print(
        "num parameters = ",
        sum(p.numel() for p in model.parameters() if p.requires_grad),
    )


# -------------------------------------------------------------------------------------------------#
class accuracy_metric:
    def __init__(self):
        self.total = 0
        self.correct = 0
        self.data = []

    def add(self, outputs, labels):
        _, predicted = torch.max(outputs.data, 1)
        self.total += labels.size(0)
        self.correct += (predicted == labels).sum().item()

    def value(self):
        val = self.correct / self.total if self.total else 0
        self.data.extend([val])
        return val


# -------------------------------------------------------------------------------------------------#
class running_mean:
    def __init__(self):
        self.total = 0
        self.count = 0

    def add(self, value):
        self.total += value
        self.count += 1

    def value(self):
        return self.total / self.count if self.count else 0


# -------------------------------------------------------------------------------------------------------#
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 3)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 3)
        self.fc1 = nn.Linear(16 * 6 * 6, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        # x = x.view(-1, 16 * 3 * 3)
        x = nn.Flatten()(x)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


# -------------------------------------------------------------------------------------------------------#
if __name__ == "__main__":

    ngpu = 1

    # CUDA for PyTorch
    if ngpu > 0 and torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        ngpu = 0
        device = torch.device("cpu")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # device=torch.device('cpu')
    print(f"Using device = {device}")
    if device.type == "cuda":
        print("GPU type =", torch.cuda.get_device_name(0))
    torch.backends.cudnn.benchmark = True

    batch_size = 100
    nworkers = 6

    # Model
    model = Net()

    if ngpu > 1:
        print("Let's use", ngpu, "GPUs!")
        model = nn.DataParallel(model)
    model.to(device)
    print(model)
    count_parameters(model)

    # Formulating generators
    training_set = DataGenerator(5000)
    validation_set = DataGenerator(1000)

    # Parameters
    params = {
        "batch_size": batch_size,  # best to use a batch_size that is a multiple of your ngpus
        "shuffle": True,
        "num_workers": nworkers,
    }

    training_generator = torch.utils.data.DataLoader(training_set, **params)
    validation_generator = torch.utils.data.DataLoader(validation_set, **params)

    # Configure
    optimizer = torch.optim.AdamW(model.parameters(), lr=0.001)

    loss_fcn = nn.CrossEntropyLoss()

    acc_list = []

    max_epochs = 5
    # ---Training and validation loop---#
    for epoch in range(max_epochs):
        # Training
        model.train()
        train_acc = accuracy_metric()
        train_loss = running_mean()
        gpu = running_mean()
        t0 = time.time()
        end = time.time()
        curr_lr = optimizer.param_groups[0]["lr"]
        for local_batch, local_labels in training_generator:
            # Transfer to GPU
            local_batch = local_batch.to(device, dtype=torch.float)
            # add labels to GPU
            local_labels = local_labels.to(device, dtype=torch.float)

            # Model computations
            # Zero the gradients
            optimizer.zero_grad()
            # forward, backward, optimize
            outputs = model(local_batch)
            loss = loss_fcn(outputs, local_labels)
            loss.backward()
            optimizer.step()

            train_loss.add(loss.item())
            train_acc.add(outputs, local_labels)
            # if(device.type =="cuda"): gpu.add(nvidia_smi.nvmlDeviceGetUtilizationRates(handle).gpu)
            end = time.time()

        # Validation
        with torch.set_grad_enabled(False):
            model.eval()
            val_loss = running_mean()
            val_acc = accuracy_metric()
            for local_batch, local_labels in validation_generator:
                # Transfer to GPU
                local_batch = local_batch.to(device, dtype=torch.float)
                # add labels to GPU
                local_labels = local_labels.to(device, dtype=torch.float)

                # local_batch, local_labels = local_batch.to(device,dtype=torch.float), local_labels.to(device,dtype=torch.float)

                # Model computations
                outputs = model(local_batch)
                loss = loss_fcn(outputs, local_labels)
                val_loss.add(loss.item())
                val_acc.add(outputs, local_labels)

        # DISPLAY PROGRESS
        # print(f"epoch={epoch:3d} dt={time.time()-t0:.3f}s gpu ={gpu.value():.0f}% training loss={train_loss.value():.5f} val loss ={val_loss.value():.5f} LR ={curr_lr:.7f} train_acc={train_acc.value():.3f} val_acc={val_acc.value():.3f} ")

        acc_list.extend([val_acc.value()])

        print(
            "completed epoch "
            + str(epoch)
            + " in "
            + "{:.3f}".format(time.time() - t0)
            + " seconds"
        )
