import torch
import torch.nn as nn

# _______________________________________________________________________________________________________
class UNet2D(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.input_tuples = config["input_tuples"]  # a list
        self.nfmaps = nfmaps = config["num_conv_filters"]  # an int
        self.inputs = config["inputs"]  # a dict
        self.scalars = config["scalar_vars"]
        self.srrf = config["scalar_reduce_res_fac"]
        nscalars = len(self.scalars)
        nclass = config["nclass"]
        bn = config["batch_norm"]
        # Notes on where/when to use BN. Suggests not to use with Dropout. Says to use before ReLU, but after Sigmoid. Use a larger LR.
        # https://machinelearningmastery.com/batch-normalization-for-training-of-deep-neural-networks/#:~:text=Batch%20normalization%20may%20be%20used,hyperbolic%20tangent%20and%20logistic%20function.
        # Paper on Group Normalization, which may be better for small batch sizes: https://arxiv.org/pdf/1803.08494.pdf

        nIn = 0
        for inp in self.input_tuples:
            nIn += inp[0]  # add number of channels

        if len(self.input_tuples) > 1:  # we have multiple resolutions
            upsamp_res = int(self.input_tuples[0][-1] / self.input_tuples[1][-1])
            self.upsample1 = nn.Upsample(scale_factor=upsamp_res, mode="nearest")
            if len(self.input_tuples) > 2:
                upsamp_res = int(self.input_tuples[0][-1] / self.input_tuples[2][-1])
                self.upsample2 = nn.Upsample(scale_factor=upsamp_res, mode="nearest")

        self.dconv_down1 = self.double_conv(nIn, nfmaps * 1, bn=bn)
        self.dconv_down2 = self.double_conv(nfmaps * 1, nfmaps * 2, bn=bn)
        self.dconv_down3 = self.double_conv(nfmaps * 2, nfmaps * 4, bn=bn)
        self.dconv_down4 = self.double_conv(nfmaps * 4, nfmaps * 8, bn=bn)

        self.dconv_down5 = self.double_conv(
            nfmaps * 8 + nscalars, nfmaps * 16 + nscalars, bn=bn
        )

        self.maxpool = nn.MaxPool2d(2, 2)
        self.maxpool4 = nn.MaxPool2d(4, 4)

        self.convtrans1 = nn.ConvTranspose2d(
            nfmaps * 16 + nscalars, nfmaps * 8, kernel_size=2, stride=2, padding=0
        )
        self.dconv_up1 = self.double_conv(nfmaps * 16, nfmaps * 8, bn=bn)

        self.convtrans2 = nn.ConvTranspose2d(
            nfmaps * 8, nfmaps * 4, kernel_size=2, stride=2, padding=0
        )
        self.dconv_up2 = self.double_conv(nfmaps * 8, nfmaps * 4, bn=bn)

        self.convtrans3 = nn.ConvTranspose2d(
            nfmaps * 4, nfmaps * 2, kernel_size=2, stride=2, padding=0
        )
        self.dconv_up3 = self.double_conv(nfmaps * 4, nfmaps * 2, bn=bn)

        self.convtrans4 = nn.ConvTranspose2d(
            nfmaps * 2, nfmaps * 1, kernel_size=2, stride=2, padding=0
        )
        self.dconv_up4 = self.double_conv(
            nfmaps * 2, nfmaps * 1, bn=bn
        )  # if you want concatenation, it would be something like this 128 +  64, 64

        # if(bn):
        #  self.conv_last = nn.Sequential(
        #                   nn.Conv2d(nfmaps, nclass, kernel_size=1, stride=1, padding=0),
        #                   nn.Sigmoid(),
        #                   nn.BatchNorm2d(nclass)
        #  )
        # else:
        self.conv_last = nn.Sequential(
            nn.Conv2d(nfmaps, nclass, kernel_size=1, stride=1, padding=0),
            # nn.Sigmoid()
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        # x is a list of inputs
        inp = x[0]
        if len(x) > 1:
            inp2 = self.upsample1(x[1])
            inp = torch.cat((inp, inp2), dim=1)
            if len(x) > 2:
                inp3 = self.upsample2(x[2])
                inp = torch.cat((inp, inp3), dim=1)

        conv1 = self.dconv_down1(inp)
        conv = self.maxpool(conv1)

        conv2 = self.dconv_down2(conv)
        conv = self.maxpool(conv2)

        conv3 = self.dconv_down3(conv)
        conv = self.maxpool(conv3)

        conv4 = self.dconv_down4(conv)
        conv = self.maxpool(conv4)

        # add 'scalars' here, at lowest resolution
        if len(self.scalars):  # scalars are last in the list, x[-1]
            conv = torch.cat((conv, x[-1]), dim=1)

        # middle
        conv = self.dconv_down5(conv)

        conv = self.convtrans1(conv)
        conv = torch.cat([conv, conv4], dim=1)
        conv = self.dconv_up1(conv)
        # x = self.upsample(x)
        # x = torch.cat([x, conv2], dim=1)

        conv = self.convtrans2(conv)
        conv = torch.cat([conv, conv3], dim=1)
        conv = self.dconv_up2(conv)

        conv = self.convtrans3(conv)
        conv = torch.cat([conv, conv2], dim=1)
        conv = self.dconv_up3(conv)

        conv = self.convtrans4(conv)
        conv = torch.cat([conv, conv1], dim=1)
        conv = self.dconv_up4(conv)

        # conv = self.conv_last(conv)

        if "CH02" in self.inputs[0]:  # reduce "500m" to "2km"
            conv = self.maxpool4(conv)
        elif "CH05" in self.inputs[0]:  # reduce "1km" to "2km"
            conv = self.maxpool(conv)

        conv = self.conv_last(conv)

        return conv

    def double_conv(self, in_channels, out_channels, ks=3, bn=False):
        P = ks // 2

        if bn:
            ngroups = self.nfmaps
            return nn.Sequential(
                nn.Conv2d(in_channels, out_channels, ks, padding=P),
                # nn.BatchNorm2d(out_channels),
                nn.GroupNorm(ngroups, out_channels),  # ngroups, nchannels
                nn.ReLU(inplace=False),
                nn.Conv2d(out_channels, out_channels, ks, padding=P),
                # nn.BatchNorm2d(out_channels),
                nn.GroupNorm(ngroups, out_channels),  # ngroups, nchannels
                nn.ReLU(inplace=False),
            )
        else:
            return nn.Sequential(
                nn.Conv2d(in_channels, out_channels, ks, padding=P),
                nn.ReLU(inplace=False),
                nn.Conv2d(out_channels, out_channels, ks, padding=P),
                nn.ReLU(inplace=False),
            )


# _______________________________________________________________________________________________________
class UNetUP(nn.Module):
    def __init__(self, config, nclass=1):
        super().__init__()
        self.input_tuples = config["input_tuples"]  # a list
        self.nfmaps = nfmaps = config["num_conv_filters"]  # an int
        self.inputs = config["inputs"]  # a dict
        self.scalars = config["scalar_vars"]
        self.srrf = config["scalar_reduce_res_fac"]
        nscalars = len(self.scalars)
        bn = config["batch_norm"]
        # Notes on where/when to use BN. Suggests not to use with Dropout. Says to use before ReLU, but after Sigmoid. Use a larger LR.
        # https://machinelearningmastery.com/batch-normalization-for-training-of-deep-neural-networks/#:~:text=Batch%20normalization%20may%20be%20used,hyperbolic%20tangent%20and%20logistic%20function.

        nIn = 0
        for inp in self.input_tuples:
            nIn += inp[0]  # add number of channels

        if len(self.input_tuples) > 1:  # we have multiple resolutions
            upsamp_res = int(self.input_tuples[0][-1] / self.input_tuples[1][-1])
            self.upsample1 = nn.Upsample(scale_factor=upsamp_res, mode="nearest")
            if len(self.input_tuples) > 2:
                upsamp_res = int(self.input_tuples[0][-1] / self.input_tuples[2][-1])
                self.upsample2 = nn.Upsample(scale_factor=upsamp_res, mode="nearest")

        self.dconv_down1 = self.double_conv(nIn, nfmaps * 1, bn=bn)
        self.dconv_down2 = self.double_conv(nfmaps * 1, nfmaps * 2, bn=bn)
        self.dconv_down3 = self.double_conv(nfmaps * 2, nfmaps * 4, bn=bn)
        self.dconv_down4 = self.double_conv(nfmaps * 4, nfmaps * 8, bn=bn)
        self.dconv_down5 = self.double_conv(nfmaps * 8 + nscalars, nfmaps * 16, bn=bn)

        self.upsample = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
        self.maxpool = nn.MaxPool2d(2, 2)
        self.maxpool4 = nn.MaxPool2d(4, 4)

        self.dconv_up1 = self.double_conv(nfmaps * 16 + nfmaps * 8, nfmaps * 8, bn=bn)
        self.dconv_up2 = self.double_conv(nfmaps * 8 + nfmaps * 4, nfmaps * 4, bn=bn)
        self.dconv_up3 = self.double_conv(nfmaps * 4 + nfmaps * 2, nfmaps * 2, bn=bn)
        self.dconv_up4 = self.double_conv(
            nfmaps * 2 + nfmaps * 1, nfmaps * 1, bn=bn
        )  # if you want concatenation, it would be something like this 128 +  64, 64

        if bn:
            self.conv_last = nn.Sequential(
                nn.Conv2d(nfmaps, nclass, kernel_size=1, stride=1, padding=0),
                nn.Sigmoid(),
                nn.BatchNorm2d(nclass),
            )
        else:
            self.conv_last = nn.Sequential(
                nn.Conv2d(nfmaps, nclass, kernel_size=1, stride=1, padding=0),
                nn.Sigmoid(),
            )

    def forward(self, x):
        # x is a list of inputs
        inp = x[0]
        if len(x) > 1:
            inp2 = self.upsample1(x[1])
            inp = torch.cat((inp, inp2), dim=1)
            if len(x) > 2:
                inp3 = self.upsample2(x[2])
                inp = torch.cat((inp, inp3), dim=1)

        conv1 = self.dconv_down1(inp)
        conv = self.maxpool(conv1)

        conv2 = self.dconv_down2(conv)
        conv = self.maxpool(conv2)

        conv3 = self.dconv_down3(conv)
        conv = self.maxpool(conv3)

        conv4 = self.dconv_down4(conv)
        conv = self.maxpool(conv4)

        # add 'scalars' here, at lowest resolution
        if len(self.scalars):  # scalars are last in the list, x[-1]
            conv = torch.cat((conv, x[-1]), dim=1)

        # middle
        conv = self.dconv_down5(conv)

        conv = self.upsample(conv)
        conv = torch.cat([conv, conv4], dim=1)
        conv = self.dconv_up1(conv)

        conv = self.upsample(conv)
        conv = torch.cat([conv, conv3], dim=1)
        conv = self.dconv_up2(conv)

        conv = self.upsample(conv)
        conv = torch.cat([conv, conv2], dim=1)
        conv = self.dconv_up3(conv)

        conv = self.upsample(conv)
        conv = torch.cat([conv, conv1], dim=1)
        conv = self.dconv_up4(conv)

        conv = self.conv_last(conv)

        if "CH02" in self.inputs[0]:  # reduce "500m" to "2km"
            conv = self.maxpool4(conv)
        elif "CH05" in self.inputs[0]:  # reduce "1km" to "2km"
            conv = self.maxpool(conv)

        return conv

    def double_conv(self, in_channels, out_channels, ks=3, bn=False):
        P = ks // 2

        if bn:
            return nn.Sequential(
                nn.Conv2d(in_channels, out_channels, ks, padding=P),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=False),
                nn.Conv2d(out_channels, out_channels, ks, padding=P),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=False),
            )
        else:
            return nn.Sequential(
                nn.Conv2d(in_channels, out_channels, ks, padding=P),
                nn.ReLU(inplace=False),
                nn.Conv2d(out_channels, out_channels, ks, padding=P),
                nn.ReLU(inplace=False),
            )


# _________________________________________________________________________________________________________________
import torch.nn.functional as F


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=False),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=False),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2), DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(
                in_channels, in_channels // 2, kernel_size=2, stride=2
            )
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        self.conv_last = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return self.conv(x)


# _______________________________________________________________________________________________________________________________
class UNet(nn.Module):
    def __init__(self, config, n_class=1):
        super(UNet, self).__init__()
        self.input_tuples = config["input_tuples"]  # a list
        self.nfmaps = nfmaps = config["num_conv_filters"]  # an int
        self.inputs = config["inputs"]  # a dict
        self.scalars = config["scalar_vars"]
        self.srrf = config["scalar_reduce_res_fac"]
        nscalars = len(self.scalars)
        bilinear = config["bilinear"]
        bn = config["batch_norm"]
        # Notes on where/when to use BN. Suggests not to use with Dropout. Says to use before ReLU, but after Sigmoid. Use a larger LR.
        # https://machinelearningmastery.com/batch-normalization-for-training-of-deep-neural-networks/#:~:text=Batch%20normalization%20may%20be%20used,hyperbolic%20tangent%20and%20logistic%20function.

        nIn = 0
        for inp in self.input_tuples:
            nIn += inp[0]  # add number of channels

        if len(self.input_tuples) > 1:  # we have multiple resolutions
            upsamp_res = int(self.input_tuples[0][-1] / self.input_tuples[1][-1])
            self.upsample1 = nn.Upsample(scale_factor=upsamp_res, mode="nearest")
            if len(self.input_tuples) > 2:
                upsamp_res = int(self.input_tuples[0][-1] / self.input_tuples[2][-1])
                self.upsample2 = nn.Upsample(scale_factor=upsamp_res, mode="nearest")

        self.inc = DoubleConv(nIn, nfmaps * 1)
        self.down1 = Down(nfmaps * 1, nfmaps * 2)
        self.down2 = Down(nfmaps * 2, nfmaps * 4)
        self.down3 = Down(nfmaps * 4, nfmaps * 8)
        factor = 2 if bilinear else 1
        self.down4 = Down(nfmaps * 8, (nfmaps * 16) // factor)
        self.up1 = Up(nfmaps * 16, (nfmaps * 8) // factor, bilinear)
        self.up2 = Up(nfmaps * 8, (nfmaps * 4) // factor, bilinear)
        self.up3 = Up(nfmaps * 4, (nfmaps * 2) // factor, bilinear)
        self.up4 = Up(nfmaps * 2, nfmaps * 1, bilinear)
        self.outc = OutConv(nfmaps * 1, n_class)
        self.maxpool = nn.MaxPool2d(2, 2)
        self.maxpool4 = nn.MaxPool2d(4, 4)

    def forward(self, x):
        # x is a list of inputs
        inp = x[0]
        if len(x) > 1:
            inp2 = self.upsample1(x[1])
            inp = torch.cat((inp, inp2), dim=1)
            if len(x) > 2:
                inp3 = self.upsample2(x[2])
                inp = torch.cat((inp, inp3), dim=1)

        x1 = self.inc(inp)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        xx = self.up1(x5, x4)
        xx = self.up2(xx, x3)
        xx = self.up3(xx, x2)
        xx = self.up4(xx, x1)
        logits = self.outc(xx)
        if "CH02" in self.inputs[0]:  # reduce "500m" to "2km"
            logits = self.maxpool4(logits)
        elif "CH05" in self.inputs[0]:  # reduce "1km" to "2km"
            logits = self.maxpool(logits)

        return logits
